"""This is the unique place in which the version number is defined.


"""

from collections import namedtuple


PRERELEASE_NORMALIZED_NAME = {"dev": "a",
                              "alpha": "a",
                              "beta": "b",
                              "candidate": "rc"}

MAJOR = 1
MINOR = 1
MICRO = 0

RELEV: str = "final"
"""RELEV must be 'dev' on the master branch, and can be "alpha", "beta",
"candidate" or "final" on a (pre)release branch."""

SERIAL: int = 0
"""SERIAL can be incremented to produce multiple alpha or release candidate
versions. It should be 0 on the development branch."""

_version_info = namedtuple(
    "version_info",
    ["major", "minor", "micro", "releaselevel", "serial"]
)

version_info = _version_info(MAJOR, MINOR, MICRO, RELEV, SERIAL)
"""Version tuple.
This can be used to reliably sort version numbers
(e.g. `if version_info[:3] <= (1, 1, 0): ...`)"""

version: str = "%d.%d.%d" % version_info[:3]
"""Version string (e.g. "1.0.0" or "1.1.0-rc0").
"""

if version_info.releaselevel != "final":
    _prerelease = PRERELEASE_NORMALIZED_NAME[version_info.releaselevel]
    version += "-%s%s" % (_prerelease, version_info.serial)



'''

Copyright (c) 2015 David Keijser

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.



Base58 encoding

Implementations of Base58 and Base58Check endcodings that are compatible
with the bitcoin network.
'''

# This module is based upon base58 snippets found scattered over many bitcoin
# tools written in python. From what I gather the original source is from a
# forum post by Gavin Andresen, so direct your praise to him.
# This module adds shiny packaging and support for python3.

__version__ = '0.2.5'

from hashlib import sha256

# 58 character alphabet used
alphabet = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'


if bytes == str:  # python2
    iseq = lambda s: map(ord, s)
    bseq = lambda s: ''.join(map(chr, s))
    buffer = lambda s: s
else:  # python3
    iseq = lambda s: s
    bseq = bytes
    buffer = lambda s: s.buffer


def b58encode_int(i, default_one=True):
    '''Encode an integer using Base58'''
    if not i and default_one:
        return alphabet[0]
    string = ""
    while i:
        i, idx = divmod(i, 58)
        string = alphabet[idx] + string
    return string


def b58encode(v):
    '''Encode a string using Base58'''
    if not isinstance(v, bytes):
        raise TypeError("a bytes-like object is required, not '%s'" %
                        type(v).__name__)

    origlen = len(v)
    v = v.lstrip(b'\0')
    newlen = len(v)

    p, acc = 1, 0
    for c in iseq(reversed(v)):
        acc += p * c
        p = p << 8

    result = b58encode_int(acc, default_one=False)

    return (alphabet[0] * (origlen - newlen) + result)


def b58decode_int(v):
    '''Decode a Base58 encoded string as an integer'''

    if not isinstance(v, str):
        v = v.decode('ascii')

    decimal = 0
    for char in v:
        decimal = decimal * 58 + alphabet.index(char)
    return decimal


def b58decode(v):
    '''Decode a Base58 encoded string'''

    if not isinstance(v, str):
        v = v.decode('ascii')

    origlen = len(v)
    v = v.lstrip(alphabet[0])
    newlen = len(v)

    acc = b58decode_int(v)

    result = []
    while acc > 0:
        acc, mod = divmod(acc, 256)
        result.append(mod)

    return (b'\0' * (origlen - newlen) + bseq(reversed(result)))


def b58encode_check(v):
    '''Encode a string using Base58 with a 4 character checksum'''

    digest = sha256(sha256(v).digest()).digest()
    return b58encode(v + digest[:4])


def b58decode_check(v):
    '''Decode and verify the checksum of a Base58 encoded string'''

    result = b58decode(v)
    result, check = result[:-4], result[-4:]
    digest = sha256(sha256(result).digest()).digest()

    if check != digest[:4]:
        raise ValueError("Invalid checksum")

    return result



CHARSET = 'qpzry9x8gf2tvdw0s3jn54khce6mua7l'


def polymod(values):
    chk = 1
    generator = [
        (0x01, 0x98f2bc8e61),
        (0x02, 0x79b76d99e2),
        (0x04, 0xf33e5fb3c4),
        (0x08, 0xae2eabe2a8),
        (0x10, 0x1e4f43e470)]
    for value in values:
        top = chk >> 35
        chk = ((chk & 0x07ffffffff) << 5) ^ value
        for i in generator:
            if top & i[0] != 0:
                chk ^= i[1]
    return chk ^ 1


def prefix_expand(prefix):
    return [ord(x) & 0x1f for x in prefix] + [0]


def calculate_checksum(prefix, payload):
    poly = polymod(prefix_expand(prefix) + payload + [0, 0, 0, 0, 0, 0, 0, 0])
    out = list()
    for i in range(8):
        out.append((poly >> 5 * (7 - i)) & 0x1f)
    return out


def verify_checksum(prefix, payload):
    return polymod(prefix_expand(prefix) + payload) == 0


def b32decode(inputs):
    out = list()
    for letter in inputs:
        out.append(CHARSET.find(letter))
    return out


def b32encode(inputs):
    out = ''
    for char_code in inputs:
        out += CHARSET[char_code]
    return out


def convertbits(data, frombits, tobits, pad=True):
    acc = 0
    bits = 0
    ret = []
    maxv = (1 << tobits) - 1
    max_acc = (1 << (frombits + tobits - 1)) - 1
    for value in data:
        if value < 0 or (value >> frombits):
            return None
        acc = ((acc << frombits) | value) & max_acc
        bits += frombits
        while bits >= tobits:
            bits -= tobits
            ret.append((acc >> bits) & maxv)
    if pad:
        if bits:
            ret.append((acc << (tobits - bits)) & maxv)
    elif bits >= frombits or ((acc << (tobits - bits)) & maxv):
        return None
    return ret



import argparse
from typing import Sequence
import warnings

from .convert import Address, guess_prefix, KNOWN_PREFIXES
from . import version


def convert():
    """This function is the entry point defined in setup.py for the command
    line ecashconvert.
    """
    warnings.warn(
        "`ecashconvert` is deprecated, use `ecashaddress convert` instead",
        DeprecationWarning
    )
    parser = argparse.ArgumentParser(description='Convert eCash address formats.')
    parser.add_argument("input_addresses", help="Input addresses to be converted.", nargs="+")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--prefix", help="Output cashaddr prefix.", default="ecash")
    group.add_argument("--legacy", help="Convert to legacy BTC address.", action="store_true")

    args = parser.parse_args()
    _convert(args.input_addresses, args.prefix, args.legacy)


def _convert(input_addresses: Sequence[str], prefix: str, is_legacy: bool):
    for addr in input_addresses:
        if is_legacy:
            print(Address.from_string(addr).to_legacy_address())
        else:
            print(Address.from_string(addr).to_cash_address(prefix))


def main():
    """This function is the entry point defined in setup.py for the command
    line ecashaddress.
    """
    parser = argparse.ArgumentParser(
        description='Tools for working with cash addresses')
    parser.add_argument("-v", "--version", help="Print the version number",
                        action="store_true")
    subparsers = parser.add_subparsers(dest='command')

    convert_parser = subparsers.add_parser(
        'convert', help = "Convert eCash address formats")
    convert_parser.add_argument(
        "input_addresses", help="Input addresses to be converted.", nargs="+")
    group = convert_parser.add_mutually_exclusive_group()
    group.add_argument("--prefix", help="Output CashAddr prefix.",
                       default="ecash")
    group.add_argument("--legacy", help="Convert to legacy BTC address.",
                       action="store_true")

    guessprefix_parser = subparsers.add_parser(
        'guessprefix',
        help=f"Guess the prefix from a CashAddr address, by trying a list of"
             f" commonly used prefixes: {', '.join(KNOWN_PREFIXES)}.")
    guessprefix_parser.add_argument(
        "address", help="Input cash address without prefix.")

    args = parser.parse_args()
    if args.version:
        print(version)
        return

    if args.command == "convert":
        _convert(args.input_addresses, args.prefix, args.legacy)
    elif args.command == "guessprefix":
        print(guess_prefix(args.address))
    elif args.command is None:
        parser.print_help()


if __name__ == '__main__':
    # This is the entry point if the package is executed via the
    # `python -m ecashaddress` command
    main()



import unittest

from .. import convert
from ..convert import Address, InvalidAddress


class TestConversion(unittest.TestCase):
    def test_to_legacy_p2sh(self):
        self.assertEqual(convert.to_legacy_address('3CWFddi6m4ndiGyKqzYvsFYagqDLPVMTzC'),
                         '3CWFddi6m4ndiGyKqzYvsFYagqDLPVMTzC')
        self.assertEqual(convert.to_legacy_address('bitcoincash:ppm2qsznhks23z7629mms6s4cwef74vcwvn0h829pq'),
                         '3CWFddi6m4ndiGyKqzYvsFYagqDLPVMTzC')

    def test_to_legacy_p2pkh(self):
        self.assertEqual(convert.to_legacy_address('155fzsEBHy9Ri2bMQ8uuuR3tv1YzcDywd4'),
                         '155fzsEBHy9Ri2bMQ8uuuR3tv1YzcDywd4')
        self.assertEqual(convert.to_legacy_address('bitcoincash:qqkv9wr69ry2p9l53lxp635va4h86wv435995w8p2h'),
                         '155fzsEBHy9Ri2bMQ8uuuR3tv1YzcDywd4')

    def test_to_cash_p2sh(self):
        self.assertEqual(convert.to_cash_address('3CWFddi6m4ndiGyKqzYvsFYagqDLPVMTzC'),
                         'ecash:ppm2qsznhks23z7629mms6s4cwef74vcwv2zrv3l8h')
        self.assertEqual(convert.to_cash_address('bitcoincash:ppm2qsznhks23z7629mms6s4cwef74vcwvn0h829pq'),
                         'bitcoincash:ppm2qsznhks23z7629mms6s4cwef74vcwvn0h829pq')

    def test_to_cash_p2pkh(self):
        self.assertEqual(convert.to_cash_address('155fzsEBHy9Ri2bMQ8uuuR3tv1YzcDywd4'),
                         'ecash:qqkv9wr69ry2p9l53lxp635va4h86wv435ugq9umvq')
        self.assertEqual(convert.to_cash_address('bitcoincash:qqkv9wr69ry2p9l53lxp635va4h86wv435995w8p2h'),
                         'bitcoincash:qqkv9wr69ry2p9l53lxp635va4h86wv435995w8p2h')

    def test_to_legacy_p2sh_testnet(self):
        self.assertEqual(convert.to_legacy_address('2MwikwR6hoVijCmr1u8UgzFMHFP6rpQyRvP'),
                         '2MwikwR6hoVijCmr1u8UgzFMHFP6rpQyRvP')
        self.assertEqual(convert.to_legacy_address('bchtest:pqc3tyspqwn95retv5k3c5w4fdq0cxvv95u36gfk00'),
                         '2MwikwR6hoVijCmr1u8UgzFMHFP6rpQyRvP')

    def test_to_legacy_p2pkh_testnet(self):
        self.assertEqual(convert.to_legacy_address('mqp7vM7eU7Vu9NPH1V7s7pPg5FFBMo6SWK'),
                         'mqp7vM7eU7Vu9NPH1V7s7pPg5FFBMo6SWK')
        self.assertEqual(convert.to_legacy_address('bchtest:qpc0qh2xc3tfzsljq79w37zx02kwvzm4gydm222qg8'),
                         'mqp7vM7eU7Vu9NPH1V7s7pPg5FFBMo6SWK')

    def test_to_cash_p2sh_testnet(self):
        self.assertEqual(convert.to_cash_address('2MwikwR6hoVijCmr1u8UgzFMHFP6rpQyRvP'),
                         'ectest:pqc3tyspqwn95retv5k3c5w4fdq0cxvv95895yhkd4')
        self.assertEqual(convert.to_cash_address('bchtest:pqc3tyspqwn95retv5k3c5w4fdq0cxvv95u36gfk00'),
                         'bchtest:pqc3tyspqwn95retv5k3c5w4fdq0cxvv95u36gfk00')

    def test_to_cash_p2pkh_testnet(self):
        self.assertEqual(convert.to_cash_address('mqp7vM7eU7Vu9NPH1V7s7pPg5FFBMo6SWK'),
                         'ectest:qpc0qh2xc3tfzsljq79w37zx02kwvzm4gyk0yx5q2a')
        self.assertEqual(convert.to_cash_address('bchtest:qpc0qh2xc3tfzsljq79w37zx02kwvzm4gydm222qg8'),
                         'bchtest:qpc0qh2xc3tfzsljq79w37zx02kwvzm4gydm222qg8')

    def test_is_valid(self):
        self.assertTrue(convert.is_valid('bitcoincash:qqkv9wr69ry2p9l53lxp635va4h86wv435995w8p2h'))
        self.assertTrue(convert.is_valid('2MwikwR6hoVijCmr1u8UgzFMHFP6rpQyRvP'))
        self.assertFalse(convert.is_valid('bitcoincash:aqkv9wr69ry2p9l53lxp635va4h86wv435995w8p2h'))
        self.assertFalse(convert.is_valid('bitcoincash:qqqqqqqq9ry2p9l53lxp635va4h86wv435995w8p2h'))
        self.assertFalse(convert.is_valid('22222wR6hoVijCmr1u8UgzFMHFP6rpQyRvP'))
        self.assertFalse(convert.is_valid(False))
        self.assertFalse(convert.is_valid('Hello World!'))

    def test_prefixes(self):
        """Test a few identical addresses with different CashAddr prefixes"""
        legacy_address = "1NLcNpAaBBMekgBZk7NxwdxwtSUTfTV8Aq"
        addr = Address.from_string(legacy_address)
        default_prefix = addr.prefix
        self.assertEqual(default_prefix, Address.MAINNET_PREFIX)

        self.assertEqual(addr.to_cash_address(prefix='ecash'),
                         'ecash:qr4pqy6q4cy2d50zpaek57nnrja7289fks00weqyz7')
        self.assertEqual(addr.to_cash_address(prefix='bitcoincash'),
                         'bitcoincash:qr4pqy6q4cy2d50zpaek57nnrja7289fkskz6jm7yf')
        self.assertEqual(addr.to_cash_address(prefix='abc'),
                         'abc:qr4pqy6q4cy2d50zpaek57nnrja7289fksqt4c50w9')
        self.assertEqual(addr.to_cash_address(prefix='simpleledger'),
                         'simpleledger:qr4pqy6q4cy2d50zpaek57nnrja7289fks6e3fw76h')

        regtest_address = 'regtest:qr4pqy6q4cy2d50zpaek57nnrja7289fksjm6es9se'
        addr2 = Address.from_string(regtest_address)
        self.assertEqual(addr2.to_legacy_address(), legacy_address)
        # The prefix defaults to the one in the input string.
        self.assertEqual(addr2.prefix, 'regtest')
        self.assertEqual(addr2.to_cash_address(), regtest_address)

    def test_prefix_case(self):
        with self.assertRaises(InvalidAddress):
            Address.from_string(
                'rEgTeSt:qr4pqy6q4cy2d50zpaek57nnrja7289fksjm6es9se')
        with self.assertRaises(InvalidAddress):
            Address.from_string(
                'regtest:QR4PQY6Q4CY2D50ZPAEK57NNRJA7289FKSJM6ES9SE')

        addr = Address.from_string('regtest:qr4pqy6q4cy2d50zpaek57nnrja7289fksjm6es9se')
        # The address should take the same case as the specified prefix
        self.assertEqual(addr.to_cash_address(prefix="SLP"),
                         'SLP:QR4PQY6Q4CY2D50ZPAEK57NNRJA7289FKSWF89PY2G')
        # Do not allow mixed-case prefixes
        with self.assertRaises(InvalidAddress):
            addr.to_cash_address(prefix="sLp")


class TestGuessPrefix(unittest.TestCase):
    def _test(self, addr, expected_prefix):
        self.assertEqual(convert.guess_prefix(addr), expected_prefix)

    def test_explicit_prefixes(self):
        self._test("ecash:qr4pqy6q4cy2d50zpaek57nnrja7289fks00weqyz7", "ecash")
        # The way address works, we always store a lower case prefix
        self._test("ECASH:QR4PQY6Q4CY2D50ZPAEK57NNRJA7289FKS00WEQYZ7", "ecash")
        self._test("foobar:qr4pqy6q4cy2d50zpaek57nnrja7289fksyz309rn7", "foobar")

    def test_fail_to_guess(self):
        # foobar:
        self._test("qr4pqy6q4cy2d50zpaek57nnrja7289fksyz309rn7", "")
        # abc:
        self._test("qr4pqy6q4cy2d50zpaek57nnrja7289fksqt4c50w9", "")

    def test_successful_guess(self):
        self._test("qr4pqy6q4cy2d50zpaek57nnrja7289fks00weqyz7", "ecash")
        self._test("QR4PQY6Q4CY2D50ZPAEK57NNRJA7289FKS00WEQYZ7", "ECASH")
        self._test("qr4pqy6q4cy2d50zpaek57nnrja7289fkskz6jm7yf", "bitcoincash")
        self._test("qr4pqy6q4cy2d50zpaek57nnrja7289fksp38mkrxf", "etoken")
        self._test("qr4pqy6q4cy2d50zpaek57nnrja7289fks6e3fw76h", "simpleledger")
        self._test("pqc3tyspqwn95retv5k3c5w4fdq0cxvv95895yhkd4", "ectest")
        self._test("pqc3tyspqwn95retv5k3c5w4fdq0cxvv95u36gfk00", "bchtest")


    def test_invalid_checksum(self):
        with self.assertRaises(InvalidAddress):
            convert.guess_prefix("ecash:qr4pqy6q4cy2d50zpaek57nnrja7289fks00000000")

    def test_mixed_case(self):
        with self.assertRaises(InvalidAddress):
            convert.guess_prefix("ECASH:qr4pqy6q4cy2d50zpaek57nnrja7289fks00weqyz7")
        with self.assertRaises(InvalidAddress):
            convert.guess_prefix("ecash:QR4PQY6Q4CY2D50ZPAEK57NNRJA7289FKS00WEQYZ7")
        with self.assertRaises(InvalidAddress):
            convert.guess_prefix("ecash:Qr4pqy6q4cy2d50zpaek57nnrja7289fks00weqyz7")


if __name__ == '__main__':
    unittest.main()



import os
import sys
from distutils.core import setup
from setuptools import find_packages

PROJECT = "ecashaddress"

def get_readme():
    """Returns content of README.rst file"""
    dirname = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(dirname, "README.rst")
    with open(filename, "r", encoding="utf-8") as fp:
        long_description = fp.read()
    return long_description


def get_version():
    """Returns current version number from _version.py file"""
    dirname = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), PROJECT)
    sys.path.insert(0, dirname)
    import _version
    sys.path = sys.path[1:]
    return _version.version


setup(name=PROJECT,
      version=get_version(),
      packages=find_packages(),
      entry_points={
          'console_scripts': [
              'ecashconvert=ecashaddress.__main__:convert',
              'ecashaddress=ecashaddress.__main__:main'
          ]

      },
      description='Python library and command line tool for converting cashaddr',
      url='https://github.com/PiRK/ecashaddress/',
      python_requires='>=3.7',
      keywords=['ecash', 'bcha', 'bitcoincash', 'bch', 'address', 'cashaddress', 'legacy', 'convert'],
      classifiers=[
          'Programming Language :: Python :: 3',
      ],
    long_description=get_readme()
)



from __future__ import annotations

from typing import Optional
import sys
import warnings

from ecashaddress.crypto import *
from ecashaddress.base58 import b58decode_check, b58encode_check


KNOWN_PREFIXES = ["ecash", "bitcoincash", "etoken", "simpleledger",
                  "ectest", "bchtest", "ecregtest", "bchreg"]


class InvalidAddress(Exception):
    pass


class Address:
    VERSION_MAP = {
        'legacy': [
            ('P2SH', 5, False),
            ('P2PKH', 0, False),
            ('P2SH-TESTNET', 196, True),
            ('P2PKH-TESTNET', 111, True)
        ],
        'cash': [
            ('P2SH', 8, False),
            ('P2PKH', 0, False),
            ('P2SH-TESTNET', 8, True),
            ('P2PKH-TESTNET', 0, True)
        ]
    }
    MAINNET_PREFIX = 'ecash'
    TESTNET_PREFIX = 'ectest'
    ALT_TESTNET_PREFIX = 'bchtest'

    def __init__(self, version, payload, prefix=None):
        self.version = version
        self.payload = payload
        if prefix:
            self.prefix = prefix
        else:
            if Address._address_type('cash', self.version)[2]:
                self.prefix = self.TESTNET_PREFIX
            else:
                self.prefix = self.MAINNET_PREFIX

    def __str__(self):
        return 'version: {}\npayload: {}\nprefix: {}'.format(self.version, self.payload, self.prefix)

    def legacy_address(self):
        warnings.warn(
            "legacy_address is deprecated, use to_legacy_address instead",
            DeprecationWarning
        )
        return self.to_legacy_address()

    def to_legacy_address(self) -> str:
        version_int = Address._address_type('legacy', self.version)[1]
        return b58encode_check(Address.code_list_to_string([version_int] + self.payload))

    def cash_address(self, prefix=None):
        warnings.warn(
            "cash_address is deprecated, use to_cash_address instead",
            DeprecationWarning
        )
        return self.to_cash_address(prefix)

    def to_cash_address(self, prefix: Optional[str] = None) -> str:
        prefix = prefix if prefix is not None else self.prefix
        self._check_case(prefix)
        is_uppercase = prefix == prefix.upper()
        version_int = Address._address_type('cash', self.version)[1]
        payload = [version_int] + self.payload
        payload = convertbits(payload, 8, 5)
        checksum = calculate_checksum(prefix, payload)
        address_string = prefix + ':' + b32encode(payload + checksum)
        if is_uppercase:
            return address_string.upper()
        return address_string

    @staticmethod
    def code_list_to_string(code_list):
        if sys.version_info > (3, 0):
            output = bytes()
            for code in code_list:
                output += bytes([code])
        else:
            output = ''
            for code in code_list:
                output += chr(code)
        return output

    @staticmethod
    def _address_type(address_type, version):
        for mapping in Address.VERSION_MAP[address_type]:
            if mapping[0] == version or mapping[1] == version:
                return mapping
        raise InvalidAddress('Could not determine address version')

    @staticmethod
    def from_string(address_string: str) -> Address:
        try:
            address_string = str(address_string)
        except Exception:
            raise InvalidAddress('Expected string as input')
        if ':' not in address_string:
            return Address.from_legacy_string(address_string)
        else:
            return Address.from_cash_string(address_string)

    @staticmethod
    def from_legacy_string(address_string: str) -> Address:
        try:
            decoded = bytearray(b58decode_check(address_string))
        except ValueError:
            raise InvalidAddress('Could not decode legacy address')
        version = Address._address_type('legacy', decoded[0])[0]
        payload = list()
        for letter in decoded[1:]:
            payload.append(letter)
        return Address(version, payload)

    @staticmethod
    def from_cash_string(address_string: str) -> Address:
        Address._check_case(address_string)
        address_string = address_string.lower()
        colon_count = address_string.count(':')
        if colon_count == 0:
            address_string = Address.MAINNET_PREFIX + ':' + address_string
        elif colon_count > 1:
            raise InvalidAddress('Cash address contains more than one colon character')
        prefix, base32string = address_string.split(':')
        decoded = b32decode(base32string)
        if not verify_checksum(prefix, decoded):
            raise InvalidAddress('Bad cash address checksum')
        converted = convertbits(decoded, 5, 8)
        version = Address._address_type('cash', converted[0])[0]
        if prefix in [Address.TESTNET_PREFIX, Address.ALT_TESTNET_PREFIX]:
            version += '-TESTNET'
        payload = converted[1:-6]
        return Address(version, payload, prefix)

    @staticmethod
    def _check_case(text):
        if text.upper() != text and text.lower() != text:
            raise InvalidAddress('Cash address contains uppercase and lowercase characters')


def to_cash_address(address):
    return Address.from_string(address).to_cash_address()


def to_legacy_address(address):
    return Address.from_string(address).to_legacy_address()


def is_valid(address):
    try:
        Address.from_string(address)
        return True
    except InvalidAddress:
        return False


def guess_prefix(cashaddress: str) -> str:
    f"""Return the lower-case prefix.

    If the prefix is not specified in the input address, a list of usual
    prefixes is tried and this function returns the first one that matches.
    The following prefixes are tried, in this order: {KNOWN_PREFIXES}.

    If prefix is specified but does not match the checksum, an InvalidAddress
    error is raised.
    If the prefix is omitted and no known prefix matches the checksum, an
    empty string is returned.
    """
    if ':' in cashaddress:
        try:
            addr = Address.from_cash_string(cashaddress)
        except InvalidAddress:
            raise
        else:
            return addr.prefix

    known_prefixes = []
    for prefix in KNOWN_PREFIXES:
        known_prefixes.append(prefix.lower())
        known_prefixes.append(prefix.upper())

    for prefix in known_prefixes:
        try:
            Address.from_cash_string(prefix + ":" + cashaddress)
        except InvalidAddress:
            pass
        else:
            return prefix
    return ""
